"""
Automatically generated file from a JSON schema.
"""


from typing import Any, TypedDict


class Configuration(TypedDict, total=False):
    """
    Tag publish configuration.

    Tag Publish configuration file (.github/publish.yaml)
    """

    transformers: "Transformers"
    """
    Transformers.

    The version transform configurations.

    default:
      pull_request_to_version:
      - to: pr-\1
    """

    docker: "Docker"
    """
    Docker.

    The configuration used to publish on Docker
    """

    pypi: "Pypi"
    """
    pypi.

    Configuration to publish on pypi
    """

    node: "Node"
    """
    node.

    Configuration to publish on node
    """

    helm: "Helm"
    """
    helm.

    Configuration to publish Helm charts on GitHub release
    """

    dispatch: list["DispatchConfig"]
    """
    Dispatch.

    default:
      []
    """



DISPATCH_CONFIG_DEFAULT: dict[str, Any] = {}
""" Default value of the field path 'Dispatch item' """



DISPATCH_DEFAULT: list[Any] = []
""" Default value of the field path 'Tag publish configuration dispatch' """



DISPATCH_EVENT_TYPE_DEFAULT = 'published'
""" Default value of the field path 'dispatch config event_type' """



DISPATCH_REPOSITORY_DEFAULT = 'camptocamp/argocd-gs-gmf-apps'
""" Default value of the field path 'dispatch config repository' """



DOCKER_AUTO_LOGIN_DEFAULT = True
""" Default value of the field path 'Docker github_oidc_login' """



DOCKER_IMAGE_GROUP_DEFAULT = 'default'
""" Default value of the field path 'Docker image group' """



DOCKER_IMAGE_TAGS_DEFAULT = ['{version}']
""" Default value of the field path 'Docker image tags' """



DOCKER_LATEST_DEFAULT = True
""" Default value of the field path 'Docker latest' """



DOCKER_REPOSITORY_DEFAULT = {'github': {'host': 'ghcr.io', 'versions_type': ['tag', 'default_branch', 'stabilization_branch', 'rebuild']}}
""" Default value of the field path 'Docker repository' """



DOCKER_REPOSITORY_VERSIONS_DEFAULT = ['tag', 'default_branch', 'stabilization_branch', 'rebuild', 'feature_branch', 'pull_request']
""" Default value of the field path 'Docker repository versions_type' """



class DispatchConfig(TypedDict, total=False):
    """
    dispatch config.

    Send a dispatch event to an other repository

    default:
      {}
    """

    repository: str
    """
    Dispatch repository.

    The repository name to be triggered

    default: camptocamp/argocd-gs-gmf-apps
    """

    event_type: str
    """
    Dispatch event type.

    The event type to be triggered

    default: published
    """



class Docker(TypedDict, total=False):
    """
    Docker.

    The configuration used to publish on Docker
    """

    latest: bool
    """
    Docker latest.

    Publish the latest version on tag latest

    default: True
    """

    images: list["DockerImage"]
    """ List of images to be published """

    repository: dict[str, "DockerRepository"]
    """
    Docker repository.

    The repository where we should publish the images

    default:
      github:
        host: ghcr.io
        versions_type:
        - tag
        - default_branch
        - stabilization_branch
        - rebuild
    """

    github_oidc_login: bool
    """
    Docker auto login.

    Auto login to the GitHub Docker registry

    default: True
    """



class DockerImage(TypedDict, total=False):
    """ Docker image. """

    group: str
    """
    Docker image group.

    The image is in the group, should be used with the --group option of tag-publish script

    default: default
    """

    name: str
    """ The image name """

    tags: list[str]
    """
    docker image tags.

    The tag name, will be formatted with the version=<the version>, the image with version=latest should be present when we call the tag-publish script

    default:
      - '{version}'
    """



class DockerRepository(TypedDict, total=False):
    """ Docker repository. """

    host: str
    """ The host of the repository URL """

    versions_type: list[str]
    """
    Docker repository versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - tag
      - default_branch
      - stabilization_branch
      - rebuild
      - feature_branch
      - pull_request
    """



HELM_PACKAGE_FOLDER_DEFAULT = '.'
""" Default value of the field path 'helm package folder' """



HELM_PACKAGE_GROUP_DEFAULT = 'default'
""" Default value of the field path 'helm package group' """



HELM_VERSIONS_DEFAULT = ['tag']
""" Default value of the field path 'helm versions_type' """



class Helm(TypedDict, total=False):
    """
    helm.

    Configuration to publish Helm charts on GitHub release
    """

    packages: list["HelmPackage"]
    """ The configuration of packages that will be published """

    versions_type: list[str]
    """
    helm versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - tag
    """



class HelmPackage(TypedDict, total=False):
    """
    helm package.

    The configuration of package that will be published
    """

    group: str
    """
    helm package group.

    The image is in the group, should be used with the --group option of tag-publish script

    default: default
    """

    folder: str
    """
    helm package folder.

    The folder of the pypi package

    default: .
    """



NODE_ARGS_DEFAULT = ['--provenance', '--access=public']
""" Default value of the field path 'node args' """



NODE_PACKAGE_FOLDER_DEFAULT = '.'
""" Default value of the field path 'node package folder' """



NODE_PACKAGE_GROUP_DEFAULT = 'default'
""" Default value of the field path 'node package group' """



NODE_REPOSITORY_DEFAULT = {'github': {'host': 'npm.pkg.github.com'}}
""" Default value of the field path 'node repository' """



NODE_VERSIONS_DEFAULT = ['tag']
""" Default value of the field path 'node versions_type' """



class Node(TypedDict, total=False):
    """
    node.

    Configuration to publish on node
    """

    packages: list["NodePackage"]
    """ The configuration of packages that will be published """

    versions_type: list[str]
    """
    node versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - tag
    """

    repository: dict[str, "NodeRepository"]
    """
    Node repository.

    The packages repository where we should publish the packages

    default:
      github:
        host: npm.pkg.github.com
    """

    args: list[str]
    """
    Node args.

    The arguments to pass to the publish command

    default:
      - --provenance
      - --access=public
    """



class NodePackage(TypedDict, total=False):
    """
    node package.

    The configuration of package that will be published
    """

    group: str
    """
    node package group.

    The image is in the group, should be used with the --group option of tag-publish script

    default: default
    """

    folder: str
    """
    node package folder.

    The folder of the node package

    default: .
    """



class NodeRepository(TypedDict, total=False):
    """ Node repository. """

    host: str
    """ The host of the repository URL """



PIP_PACKAGE_GROUP_DEFAULT = 'default'
""" Default value of the field path 'pypi package group' """



PYPI_PACKAGE_FOLDER_DEFAULT = '.'
""" Default value of the field path 'pypi package folder' """



PYPI_VERSIONS_DEFAULT = ['tag']
""" Default value of the field path 'pypi versions_type' """



class Pypi(TypedDict, total=False):
    """
    pypi.

    Configuration to publish on pypi
    """

    packages: list["PypiPackage"]
    """ The configuration of packages that will be published """

    versions_type: list[str]
    """
    pypi versions.

    The kind or version that should be published, tag, branch or value of the --version argument of the tag-publish script

    default:
      - tag
    """



class PypiPackage(TypedDict, total=False):
    """
    pypi package.

    The configuration of package that will be published
    """

    group: str
    """
    pip package group.

    The image is in the group, should be used with the --group option of tag-publish script

    default: default
    """

    folder: str
    """
    pypi package folder.

    The folder of the pypi package

    default: .
    """

    build_command: list[str]
    """ The command used to do the build """



TRANSFORMERS_DEFAULT = {'pull_request_to_version': [{'to': 'pr-\\1'}]}
""" Default value of the field path 'Tag publish configuration transformers' """



TRANSFORM_DEFAULT: list[Any] = []
""" Default value of the field path 'transform' """



TRANSFORM_FROM_DEFAULT = '(.+)'
""" Default value of the field path 'Version transform from_re' """



TRANSFORM_TO_DEFAULT = '\\1'
""" Default value of the field path 'Version transform to' """



Transform = list["VersionTransform"]
"""
transform.

A version transformer definition

default:
  []
"""



class Transformers(TypedDict, total=False):
    """
    Transformers.

    The version transform configurations.

    default:
      pull_request_to_version:
      - to: pr-\1
    """

    branch_to_version: "Transform"
    """
    transform.

    A version transformer definition

    default:
      []
    """

    tag_to_version: "Transform"
    """
    transform.

    A version transformer definition

    default:
      []
    """

    pull_request_to_version: "Transform"
    """
    transform.

    A version transformer definition

    default:
      []
    """



class VersionTransform(TypedDict, total=False):
    """ Version transform. """

    from_re: str
    """
    transform from.

    The from regular expression

    default: (.+)
    """

    to: str
    """
    transform to.

    The expand regular expression: https://docs.python.org/3/library/re.html#re.Match.expand

    default: \1
    """




"""Some utility functions for Docker images."""

import subprocess  # nosec: B404
from pathlib import Path
from typing import Optional, cast

import yaml
from debian_inspector.version import Version


def get_dpkg_packages_versions(
    image: str,
    default_distribution: Optional[str] = None,
    default_release: Optional[str] = None,
) -> tuple[bool, dict[str, Version]]:
    """
    Get the versions of the dpkg packages installed in the image.

    `get_dpkg_packages_versions("org/image:tag")` will return something like:
    (true, {"debian_11/api": "2.2.0", ...})

    Where `debian_11` corresponds on last path element for 'Debian 11'
    from https://repology.org/repositories/statistics
    """
    os_release = {}
    try:
        os_release_process = subprocess.run(
            ["docker", "run", "--rm", "--entrypoint=", image, "cat", "/etc/os-release"],
            stdout=subprocess.PIPE,
            check=True,
        )
        os_release = dict([e.split("=") for e in os_release_process.stdout.decode().split("\n") if e])
    except subprocess.CalledProcessError:
        print("Info: /etc/os-release not found in the image")

    lsb_release = {}
    try:
        lsb_release_process = subprocess.run(
            ["docker", "run", "--rm", "--entrypoint=", image, "cat", "/etc/lsb-release"],
            stdout=subprocess.PIPE,
            check=True,
        )
        lsb_release = dict([e.split("=") for e in lsb_release_process.stdout.decode().split("\n") if e])
    except subprocess.CalledProcessError:
        print("Info: /etc/lsb-release not found in the image")

    distribution = os_release.get("ID", lsb_release.get("DISTRIB_ID", default_distribution))
    release = os_release.get("VERSION_ID", lsb_release.get("DISTRIB_RELEASE", default_release))
    if distribution is None:
        print("Could not get the distribution of the image, you should provide a default distribution")
        return False, {}
    if release is None:
        print("Could not get the release of the image, you should provide a default release")
        return False, {}

    distribution_final = distribution.strip('"').lower()
    release_final = release.strip('"').replace(".", "_")
    prefix = f"{distribution_final}_{release_final}/"
    print(f"Found distribution '{distribution_final}', release '{release_final}'.")

    if distribution_final == "ubuntu" and release_final == "18_04":
        print("Warning: Ubuntu 18.04 is not supported")
        return False, {}

    package_version: dict[str, Version] = {}
    packages_status_process = subprocess.run(
        ["docker", "run", "--rm", "--entrypoint=", image, "dpkg", "--status"],
        stdout=subprocess.PIPE,
        check=True,
    )
    packages_status_1 = packages_status_process.stdout.decode().split("\n")
    packages_status_2 = [e.split(": ", maxsplit=1) for e in packages_status_1]
    packages_status = [e for e in packages_status_2 if len(e) == 2]
    package = None
    version = None
    for name, value in packages_status:
        if name == "Package":
            if package is not None:
                if version is None:
                    print(f"Error: Missing version for package {package}")
                else:
                    if package in package_version and version != package_version[package]:
                        print(
                            f"The package {package} has different version "
                            f"({package_version[package]} != {version})",
                        )
                    if package not in ("base-files",):
                        package_version[package] = version
            package = value
            version = None
        if name == "Version" and version is None:
            version = Version.from_string(value)

    return True, {f"{prefix}{k}": v for k, v in package_version.items()}


def get_versions_config() -> tuple[dict[str, dict[str, str]], bool]:
    """Get the versions from the config file."""
    dpkg_versions_path = Path(".github/dpkg-versions.yaml")
    if dpkg_versions_path.exists():
        with dpkg_versions_path.open(encoding="utf-8") as versions_file:
            return (
                cast(dict[str, dict[str, str]], yaml.load(versions_file.read(), Loader=yaml.SafeLoader)),
                True,
            )
    return {}, False


def check_versions(
    versions_config: dict[str, str],
    image: str,
    default_distribution: Optional[str] = None,
    default_release: Optional[str] = None,
) -> bool:
    """
    Check if the versions are correct.

    The versions of packages in the image should be present in the config file.
    The versions of packages in the image shouldn't be older than the versions of the config file.
    """
    result, versions_image = get_dpkg_packages_versions(image, default_distribution, default_release)
    if not result:
        return False

    success = True
    for package, version in versions_image.items():
        if package not in versions_config:
            print(f"Package {package} with version {version} is not in the config file for the image {image}")
            success = False
        elif Version.from_string(versions_config[package]) > version:
            print(
                f"Package {package} is older than the config file for the image {image}: "
                f"{versions_config[package]} > {version}.",
            )
            success = False

    return success
